/*
 * Copyright ©Rastislav Mirek. All rights reserved.
 * No waranty for this software or derived works.
 */

package cz.muni.fi.pb162.trainexample;

/**
 * This class represents Train.
 *
 * Created 12.11.2009 at 18:39:06.
 *
 * @author Rastislav Mirek <rmirek at mail.muni.cz>
 * @version %I%, %G%
 * @since 1.0
 */
public class Train {

    private Locomotive locomotive;
    private Carriage[] carriages;
    private int carriageCount = 0; // holds the last empty slot for carriage

    public static final int POWER_FOR_CARRIAGE = 50; //in kW
    public static final int MAX_CARRIAGE_IN_TRAIN = 30;
    public static final int MAX_TRAIN_LENGTH_WITHOUT_LOCOMOTIVE = 500; //in meters

    /**
     * Constructs new train
     * @param locomotive not null locomotive object that will be a locomotive for this train. Max number of carriages is computed according to locomotives power
     * @throws NullPointerException if locomotive is null
     */
    public Train(Locomotive locomotive) {
        if(locomotive == null){
            throw new NullPointerException("locomotive");
        }
        this.locomotive = locomotive;

        int maxNumberOfCarraiges = Math.min(MAX_CARRIAGE_IN_TRAIN, (int)(locomotive.getPower() / POWER_FOR_CARRIAGE));
        carriages = new Carriage[maxNumberOfCarraiges];
    }

    /**
     * Adds given carriage to the train
     *
     * NOTE: demostrates use of checked exceptions. Possible throwing of checked
     * exception CannotAddCarriageException has to be declared in method header.
     * There is no need to declare TooLongTrainException in the same way as it is only
     * special case of CannotAddCarriageException
     *
     * @param carriage not null carriadge to add
     * @throws CannotAddCarriageException when trying to add carriage to already full train
     * @throws NullPointerException if carriage is null
     * @throws TooLongTrainException if the train after adding given carriage would be longer then allowed
     */
    public void addCarriage(Carriage carriage) throws CannotAddCarriageException{
        if(carriage == null){
            throw new NullPointerException("carriage");
        }
        if(carriageCount + 1 > carriages.length){
            throw new CannotAddCarriageException("Cannot add carriage " + carriage + " to train " + this +
                    " because max carriage number would be exeded. Train can have maximum " + carriages.length +
                    " carriages");
        }

        if(getLengthWithoutLocomotive() > MAX_TRAIN_LENGTH_WITHOUT_LOCOMOTIVE){
            throw new TooLongTrainException("Cannot add carriage " + carriage + " to train " + this +
                    " because train would be to long. The train would have " + getLengthWithoutLocomotive() +
                    " meters.");
        }

        carriages[carriageCount] = carriage;
        carriageCount++;
    }

    /**
     * Computes length of this train not counting locomotive
     * @return length of this train not counting locomotive in meters
     */
    public double getLengthWithoutLocomotive(){
        double trainLength = 0;
        for(int i = 0; i < carriageCount; i++){
            trainLength = trainLength + carriages[i].getLength();
        }
        return trainLength;
    }

    /**
     * Standard getter for locomotive field
     * @return locomotive of this train
     */
    public Locomotive getLocomotive() {
        return locomotive;
    }

    /**
     * Standard getter for carriage count field
     * @return carriage count of this train
     */
    public int getCarriageCount() {
        return carriageCount;
    }

    /**
     * Gets specified carriage in the train.
     *
     * NOTE: demonstrates use of checked exceptions. See documentation to method {@link Train#addCarriage(cz.muni.fi.pb162.trainexample.Carriage) for details on similar demonstration
     * 
     * @param carriageNumber number specifying which Carriage to return. Carriages in the train are indexed from locomotive. First carriage afret locomotive has number 0, second number 1,...
     * @return carriage at given possition from locomotive
     * @throws NoCarriageException if carriageNumber >= {@link Train#getCarriageCount()}
     */
    public Carriage getCarriage(int carriageNumber) throws NoCarriageException{
        if(carriageNumber < 0){
            //here one could throw also ArrayIndexOutOfException
            throw new IllegalArgumentException("No train has negative carriages :)." +
                    "Carriages are indexed from zero, you have query " + carriageNumber + "th carriage");
        }
        if(carriageNumber >= carriageCount){
            //this may be better with IllegalArgumentException or ArrayIndexOutOfException
            //because they are unchecked. Checked exception is used here to demonstrate
            //checked Exception feature of Java language
            throw new NoCarriageException("Such carriage was not added yet. Train has only " + carriageCount +
                    " cariages. You have query " + carriageNumber + "th carriage and they are indexed from 0");
        }

        return carriages[carriageNumber];
    }

    /**
     * Computes amount of cargo in the train in tons.
     *
     * NOTE: Demonstrates use of key word <code>instanceof</code> and type changes
     *
     * @return amount of cargo in the train in tons
     */
    public double getCargoCount(){
        double cargoCount = 0;
        for(int i = 0; i < carriageCount; i++){
            //if Carriage at position i is instance of CargoCarriage
            if(carriages[i] instanceof CargoCarriage){
                //change type of i-th Carriage to CargoCariage
                CargoCarriage ithCargoCarriage = (CargoCarriage) carriages[i];
                cargoCount = cargoCount + ithCargoCarriage.getActualCargo();
            }
        }
        return cargoCount;
    }

    /**
     * Computes number of passanger currently in the train
     *
     * NOTE: Demonstrates use of key word <code>instanceof</code> and type changes
     *
     * @return number of passangers in train
     */
    public int getPassangersCount(){
        int passangersCount = 0;
        for(int i = 0; i < carriageCount; i++){
            //if Carriage at position i is instance of PersonalCarriage
            if(carriages[i] instanceof PersonalCarriage){
                //change type of i-th Carriage to PersonalCarriage
                PersonalCarriage ithPersonalCarriage = (PersonalCarriage) carriages[i];
                passangersCount = passangersCount + ithPersonalCarriage.getPassangerCount();
            }
        }
        return passangersCount;
    }

    /**
     * Holds overall contract of method equals (see {@link Object#equals(java.lang.Object)}).
     * Given object is equal to this Train if and only if is Train, has equal locomotive to
     * locomotive of this train and has same number of carriages.
     *
     * NOTE: demostrates proper implementation of equals. When you override equals you must always
     * override {@link Object#hashCode()} too. Uses key word <code>instanceof</code> and type changing.
     * For other examples on this see {@link Train#getPassangersCount()} and {@link Train#getCargoCount()}
     * methods.
     *
     * @param obj object in task
     * @return true if and only if this train is equal to given object
     */
    @Override
    public boolean equals(Object obj) {
        if (obj == null) {
            return false;
        }
        //check if obj is instance of Train
        if(!(obj instanceof Train)){
            //if obj is not instance of Train it is not equal to this Train. Return false
            return false;
        }

        //obj is instance of Train. Change its type to Train
        Train other = (Train) obj;
        //direct access to atributes of other object!!! Recomended only in this case(method equals).
        return locomotive.equals(other.locomotive) && carriageCount == other.carriageCount;
    }

    /**
     * Return the number that is always the same for equal trains,
     * in most cases schoul be different for not equals trains but sometimes may be same.
     * That is in the other words contract of method hashCode for this class.
     * For more details on overall constract see {@link Object#hashCode()}
     *
     * @return hash code value for this object.
     */
    @Override
    public int hashCode() {
        return 3 * locomotive.hashCode() + 5 * carriageCount;
    }

    @Override
    public String toString() {
        String result = "Train with locomotive " + locomotive + " and " + carriageCount + " carriages.";
        for(int i = 0; i < carriageCount; i++){
            result = result + i + "th carriage is " + carriages[i] + ".";
        }
        return result;
    }







}
